#ifndef _Memory_Block__H
#define _Memory_Block__H

#include "Memory.h"
#include "Memory_Block.h"
#include "Prefix.h"
#include "EARTH_PT.h"
#include "windrvr.h"

namespace EARTH {
namespace Memory {
namespace KERNEL_BIT {
	class Block_ : public Block {
	public:
		Block_(void *driver, uint card, uint offset, uint size) : mDriver(driver), mCard(card), mOffset(offset), mSize(size), mPage(Page(offset, size))
		{
			mPtr = NULL;
			mDMA = NULL;
			mIsLocked = false;

			_ASSERTE(mSize);
		}

		virtual ~Block_()
		{
		}

		virtual void SetBasePtr(void *ptr)
		{
			_ASSERTE(mPtr == NULL);
			mPtr = static_cast<byte *>(ptr);
			_ASSERTE(mPtr);
		}

		virtual void *Ptr() const
		{
			_ASSERTE(mPtr);
			void *ptr = mPtr + mOffset;
			return ptr;
		}

		virtual uint Size() const
		{
			_ASSERTE(mSize);
			return mSize;
		}

		virtual status Lock(Direction direction)
		{
			_ASSERTE(mIsLocked == false);
			if (mIsLocked) { _ASSERTE(false); return PT::STATUS_GENERAL_ERROR; }

			uint dma_size = DMASize();
			WD_DMA *dma = static_cast<WD_DMA *>(::malloc(dma_size));
			if (dma == NULL) { _ASSERTE(false); return PT::STATUS_GENERAL_ERROR; }
			::memset(dma, 0, dma_size);

			uint option = 0;
			if (WD_DMA_PAGES < mPage             ) option |= DMA_LARGE_BUFFER;
			if (direction & DIRECTION_FROM_DEVICE) option |= DMA_FROM_DEVICE;
			if (direction & DIRECTION_TO_DEVICE  ) option |= DMA_TO_DEVICE;

			#if defined(KERNEL_64BIT)
			option |= DMA_ALLOW_64BIT_ADDRESS;
			#endif

			void *ptr = Ptr();
			_ASSERTE(ptr);
			dma->pUserAddr = ptr;
			dma->dwBytes   = mSize;
			dma->dwOptions = option;
			dma->dwPages   = mPage;
			dma->hCard     = mCard;

			_ASSERTE(mDriver != INVALID_HANDLE_VALUE);
			DWORD status = WD_DMALock(mDriver, dma);
			if (status != WD_STATUS_SUCCESS) {
				::free(dma);
				dma = NULL;

				return PT::STATUS_WD_DMALock_ERROR;
			}

			_ASSERTE(mDMA == NULL);
			mDMA = dma;

			mIsLocked = true;

			return status;
		}

		virtual status Unlock()
		{
			_ASSERTE(mIsLocked);
			if (mIsLocked == false) { _ASSERTE(false); return PT::STATUS_GENERAL_ERROR; }

			if (mDMA) {
				_ASSERTE(mDriver != INVALID_HANDLE_VALUE);
				DWORD status = WD_DMAUnlock(mDriver, mDMA);

				::free(mDMA);
				mDMA = NULL;

				if (status != WD_STATUS_SUCCESS) { _ASSERTE(false); return PT::STATUS_WD_DMAUnlock_ERROR; }
			}

			mIsLocked = false;

			return PT::STATUS_OK;
		}

		virtual bool IsLocked() const
		{
			return mIsLocked;
		}

		virtual void GetPhysicalAddress(uint *infoCount, const PhysicalAddressInfo **physicalAddressInfo) const
		{
			_ASSERTE(mDMA);
			_ASSERTE(infoCount); *infoCount = mDMA->dwPages;
			_ASSERTE(physicalAddressInfo); *physicalAddressInfo = reinterpret_cast<const PhysicalAddressInfo *>(mDMA->Page);
		}

		virtual status Sync(sync sync)
		{
			_ASSERTE(mIsLocked);
			if (mIsLocked == false) { _ASSERTE(false); return PT::STATUS_WD_DMAUnlock_ERROR; }

			DWORD status;

			_ASSERTE(mDriver != INVALID_HANDLE_VALUE);
			_ASSERTE(mDMA);
			switch (sync) {
				case SYNC_CPU: status = WD_DMASyncCpu(mDriver, mDMA);                  break;
				case SYNC_IO : status = WD_DMASyncIo (mDriver, mDMA);                  break;
				default      : status = WD_STATUS_SUCCESS;            _ASSERTE(false); break;
			}

			if (status != WD_STATUS_SUCCESS) {
				switch (sync) {
					case SYNC_CPU: return PT::STATUS_WD_DMASyncCpu_ERROR;
					case SYNC_IO : return PT::STATUS_WD_DMASyncIo_ERROR;
					default      : _ASSERT(false); break;
				}
			}
	
			return PT::STATUS_OK;
		}

	protected:
		const HANDLE mDriver;
		const DWORD mCard;
		const uint mOffset;
		const uint mSize;
		const uint mPage;
	
		byte *mPtr;
		WD_DMA *mDMA;
		bool mIsLocked;

		uint Page(uint offset, uint size) const
		{
			_ASSERTE((offset % SIZE) == 0);
			uint page = (size + SIZE-1) / SIZE;

			return page;
		}

		uint DMASize() const
		{
			uint page;

			if (mPage <= WD_DMA_PAGES) {
				page = 0;
			} else {
				page = mPage - WD_DMA_PAGES;
			}

			uint size = sizeof(WD_DMA) + sizeof(WD_DMA_PAGE)*page;

			return size;
		}
	};

	Block *NewBlock(void *driver, uint card, uint offset, uint size)
	{
		Block *block = new Block_(driver, card, offset, size);
		return block;
	}
}
}
}

#endif
